;; Programa: Package con agentes jugadores de PlanTron basados
;;           en búsquedas con sistemas de producción que 
;;           impelmentan depredadores
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (Sistemas de Producción)
;; Autor:    Patricio García (patricio AT etsii.ull.es)
;; Fecha:    13/04/2009
;; Comentarios:
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE Y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;           ....
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;;           Hace uso del paquete PT-UTIL con agentes pasivos
;;           con distintas utilidades para asistir a agentes jugadores
;;           de PlanTron (agente pasivo CONTADOR) e interfaces para
;;           acceso a paquetes SHELL, GENERIC-SEACH y A_STAR.
;;
;;           Hace uso del paquete interfaz para la creación de sistemas
;;           de producción simples (SHELL) a través de PT-UTIL.
;;
;;           Hace uso del paquete marco genérico para la búsquedas en
;;           sistemas de producción simples (GENERIC-SEARCH) a través
;;           de PT-UTIL.
;;
;;           Hace uso del paquete para la búsquedas A* (A_STAR) en
;;           sistemas de producción simples a través de PT-UTIL
;;
;;           Hace uso inderecto del paquete PT-VIS mediante la
;;           solicitud de impresión del camino encontrado en la
;;           función PRINT-CAMINO.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (13/04/2009): Primera versión basada en 
;;             agentes-aluXXXX-camino-2008.lisp
;;
;;           - Ver 0.1.2 (20/04/2009): Correcciones menores
;;
;;
;; Por hacer:
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-AGENTES-GENDIR"
  (:use "COMMON-LISP")
  (:documentation "Agentes jugadores de PlanTron generadores de direcciones basado en búsquedas.")
  (:export "VERSION"))

(in-package "PT-AGENTES-GENDIR")


;;
;; Constante versión
;; -----------------

(defconstant version "0.1.2"
  "Versión del package PT-AGENTES-GENDIR.")


;;
;; Definición de estructura de estados
;; -----------------------------------

(defstruct estado
  "Estructura para almacenar los estados del sistema de producción"
  (camino nil)       ; Camino como lista inversa de posiciones (x y)
  (lon-camino 0)     ; Longitud del camino
  (f-heur 0)         ; Valor de la función heurística f'
  (dir-ini nil))     ; Dirección inicial del camino: [nil, 0, 1, 2 ó 3]


;;
;; Definición de varialbes globales
;; --------------------------------

(defvar *f-heur* nil
  "Función heurística que evalua un estado en acción de regla.")

(defvar *objetivo-p* nil
  "Función estado objetivo que detecta condición de parada")

(defvar id-dpdd nil
  "Identificador del depredador")

(defvar num-jug-gdir nil
  "Número de jugadores tipo gdir")

(defvar mej-est nil
  "Mejor estado encontrado")

(defvar lon-mej-est nil
  "Longitud del camino del mejor estado encontrado")


;;
;; Funciones de apoyo para elaboración de reglas
;; ---------------------------------------------

(defun no-obs-ni-dpdd (cont)
  "Devuelve T si no es ni obstáculo ni el propio depredador"
  (not (or (eq cont id-dpdd)
           (eq (envio:consulta-tipo cont) 'envio:obstaculo))))

(defun pos-posible-p (pos &optional dir)
  "Mira si es posible colocarse (sin obstáculo o rastro del propio agente) en la posición resultante de moverse en la dirección DIR a la posición POS. Devuelve T en tal caso y NIL en caso contrario. POS es una posición del tablero dada por una lista con valor de coordenada X y valor de coordenada Y. DIR ha de ser un número entero (dirección) entre 0 y 3 (norte, este, sur y oeste respectivamente) o NIL si se quiere consultar la posición POS."
  (no-obs-ni-dpdd (envio:consulta-pos-tab pos dir)))

(defconstant dir-inc
  (make-array '(4 2)
	      :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de dirección a número de coordenada a modificar y valor de incremento.")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defun pos-inc-dir (pos dir)
  "Calcula nueva lista con la posición resultante de desplazarse en la dirección DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))


;;
;; Funciones para generar sistema de producción
;; --------------------------------------------

(defun estados-iniciales (id)
  "Genera lista de estados iniciales del jugador con identificador ID."
  (let ((nue-est (make-estado :camino (list (envio:consulta-pos-jug id)))))
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    (list nue-est)))

(defun estado-objetivo-p (est)
  "Chequea si un estado EST es estado objetivo o NIL."
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      (let ((dest (envio:consulta-pos-tab (first (estado-camino est)))))
	(setf mej-est est)
	(and dest                       ; o hay algo
	     (> dest id-dpdd)           ;   y con id mayor que dpdd
             (<= dest num-jug-gdir))))) ;   y que no sea una planta

(defun precond-regla-dir (est dir)
  "Generador de precondicion de regla de mover el estado EST en dirección DIR."
  (let ((pos-est (first (estado-camino est))))
    ; La posición no pertenece al propio camino
    (and (not (member (pos-inc-dir pos-est dir)
		      (rest (estado-camino est))
		      :test #'equal))
	 ; Y la posición se puede pisar
	 (pos-posible-p pos-est dir))))

(defun accion-regla-dir (est dir)
  "Generador de acción de regla de mover el estado EST en dirección DIR."
  (let ((nue-est (make-estado :camino 
			      (cons (pos-inc-dir (first (estado-camino est)) dir)
				    (estado-camino est))
			      :lon-camino (1+ (estado-lon-camino est))
			      :dir-ini (or (estado-dir-ini est) dir))))
    ; Asignar valor heurísticos, si se requiere
    (if *f-heur*
	(setf (estado-f-heur nue-est) (funcall *f-heur* nue-est)))
    nue-est))


;;
;; Funciones de apoyo para agentes de búsqueda de camino
;; -----------------------------------------------------

(defun print-camino (camino)
  "Solicita imprimir la lista de posiciones dada en CAMNIO a los agentes capaces de acpetar solicitudes tipo :punteo."
  (envio:broadcast :punteo (reverse camino)))


;;
;; Inicializador de agentes de búsqueda de camino
;; ----------------------------------------------

(defun dpdd-ini (id tipo-juego)
  "Inicialización genérica de agente que busca camino."
  ; Inicializar var globales
  (setf id-dpdd id)
  (setf num-jug-gdir (getf tipo-juego :num-jug-gdir))
  (if (pt-util:usa-sistema 'camino-dpdd)
      ; Creamos nuevo sistema de producción
      (progn
	; Insertar las nuevas reglas
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 3))  ; Regla Oeste
		       #'(lambda (est) (accion-regla-dir est 3)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 1))  ; Regla Este
		       #'(lambda (est) (accion-regla-dir est 1)))
	(pt-util:regla #'(lambda (est) (precond-regla-dir est 0))  ; Regla Norte
		       #'(lambda (est) (accion-regla-dir est 0)))
      (pt-util:regla #'(lambda (est) (precond-regla-dir est 2))    ; Regla Sur
		     #'(lambda (est) (accion-regla-dir est 2)))))
  ; Cargar heurística a utilizar (sin heurística)
  (setf *f-heur* nil)
  t)


;;
;; Agente búsqueda de camino en anchura
;; ------------------------------------

(defun dpddba-ini (id tipo-juego)
  "Inicialización de agente que busca camino en anchura."
  ; Llamar a inicialización de agentes camino
  (dpdd-ini id tipo-juego)
  (setf *objetivo-p* #'estado-objetivo-p))

(defun dpddba (id)
  "Agente jugador PlanTron que busca camino a otro jugador gendir moviéndose en anchura."
  ; Iniciar contador de tiempo
  (pt-util:inicia-contador)
  ; Inicializa mejor estado
  (setf mej-est nil)
  ; Usar sistema de producción del depredador
  (pt-util:usa-sistema 'camino-dpdd)
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  *objetivo-p*
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'(lambda (new-states)
			  (nconc pt-util:*search-queue* new-states))
	  :dequeue-fn #'(lambda () (pop pt-util:*search-queue*)))))
    ; Preguntar por condiciones de parada
    (cond
     ((null est-sol)              ; No encontrada solución (encerrado)
      (envio:mensaje "Encerrado :(")
      (estado-dir-ini mej-est))                ; Mover según mejor estado
     
     ((pt-util:consulta-contador) ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (estado-dir-ini est-sol))                ; Mover según primero frente
     
     (t                           ; Encontrado camino
      (envio:mensaje "Encontrado camino")
      (print-camino (estado-camino est-sol))    ; Imprimir camino hallado
      (estado-dir-ini est-sol)))))              ; Mover según primero frente


;;
;; Funciones para agentes informados
;; ---------------------------------

(defun distancia-city (pos1 pos2)
  "Calcula la distancia city block entre POS1 y POS2"
  (+ (abs (- (first pos1) (first pos2)))
     (abs (- (second pos1) (second pos2)))))

(defun dist-gdir-cerca (pos)
  "Calcula la distancia city-block desde la posición POS al jugador gdir más cercano a dicha posición. Si no hay gdirs devuelve NIL."
  (let (dist
        posgdir)
    ; Los identificadores de gdir a buscar son mayores que id-dpdd (presas)
    (dotimes (i (- num-jug-gdir id-dpdd) dist)
      (setf posgdir (envio:consulta-pos-jug (+ i id-dpdd 1)))
      (if posgdir
	  (if dist
	      (setf dist (min dist (distancia-city posgdir pos)))
	    (setf dist (distancia-city posgdir pos)))))))


;;
;; Agente búsqueda primero el mejor y A*
;; -------------------------------------

(defun h-dist-gdir (est)
  "Función heurística subdividida con coste del camino acumulado + distancia city block que falta hasta el agente gdir mas cercano"
      ; Si localiza gdir estima la distancia a él
  (or (dist-gdir-cerca (first (estado-camino est))) 
	   0))    ; sino supone ya pegado

(defun f-dist-gdir (est)
  "Función heurística subdividida con coste del camino acumulado + distancia city block que falta hasta el agente gdir mas cercano"
  (let ((dt (dist-gdir-cerca (first (estado-camino est)))))
    (+ (estado-lon-camino est)
       (or dt        ; Si localiza gdir estima la distancia a él
	   0))))     ; sino supone ya pegado

(defun estado-objetivo-bf-p (est)
  "Chequea si un estado EST es estado objetivo o NIL. Aprovecha el campo F-HEUR del estado."
  ; Actualizar mejor estado
  (if (and est 
	   (or (not mej-est) (> (estado-lon-camino est) (estado-lon-camino mej-est))))
      (setf mej-est est
	    lon-mej-est (estado-lon-camino est)))
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      (zerop (estado-f-heur est))))  ; o ya ha llegado

(defun estado-objetivo-a-p (est)
  "Chequea si un estado EST es estado objetivo o NIL. Aprovecha el campo F-HEUR del estado."
  ; Actualizar mejor estado
  (if (and est
	   (or (not mej-est) (> (estado-lon-camino est) (estado-lon-camino mej-est))))
      (setf mej-est est
	    lon-mej-est (estado-lon-camino est)))
  ; Si se ha acabado el tiempo o no se ha encontrado solución
  (or (pt-util:consulta-contador) (null est)
      (= (estado-lon-camino est) (estado-f-heur est))))  ; o ya ha llegado

(defun igual-estado (est1 est2)
  "Comprueba si el estado EST1 es igual al EST2"
  ; Considera estados iguales si su camino acaba en la misma posición
  (equal (first (estado-camino est1))
         (first (estado-camino est2))))

(defun dpddbf-ini (id tipo-juego)
  "Inicialización de agente que busca primero el mejor."
  ; Llamar a inicialización de agentes camino
  (dpdd-ini id tipo-juego)
  ; Inicializar variables para cálculo de heurística
  (setf *f-heur* #'h-dist-gdir)                       ; Fun. f' para reglas
  ; Inicializar variables para cálculo de estado objetivo
  (setf *objetivo-p* #'estado-objetivo-bf-p)          ; Fun. estado objetivo
  t)

(defun dpdda-ini (id tipo-juego)
  "Inicialización de agente que busca camino con A*."
  (dpddbf-ini id tipo-juego)
  (setf *f-heur* #'f-dist-gdir)                        ; Fun. f' para reglas
  (setf *objetivo-p* #'estado-objetivo-a-p))

(defun dpdda (id)
  "Agente jugador PlanTron que busca camino a otro jugador gendir moviéndose con algoritmo A."
  ; Iniciar contador de tiempo
  (pt-util:inicia-contador)
  ; Inicializa mejor estado
  (setf mej-est nil)
  ; Usar sistema de producción de id
  (pt-util:usa-sistema 'camino-dpdd)
  ; Inicializar variables para heurísticas del A*
  (setf pt-util:*coste-acum-fn* #'estado-lon-camino)  ; Fun. g  para búsqueda
  (setf pt-util:*eval-fn* #'estado-f-heur)            ; Fun. h' para búsqueda
  (setf pt-util:*igual-est-fn* #'igual-estado)        ; Pred. estados iguales
  ; Lanzar búsqueda
  (let ((est-sol
	 (pt-util:generic-search
	  (estados-iniciales id)
	  *objetivo-p*
	  #'pt-util:hijos
	  :display-fn #'null
	  :merge-fn   #'pt-util:priority-a-merge-fn
	  :dequeue-fn #'pt-util:priority-a-dequeue-fn)))
    ; Preguntar por condiciones de parada
    (cond
     ((null est-sol)              ; No encontrada solución (encerrado)
      (envio:mensaje "Encerrado :(")
      (estado-dir-ini mej-est))                ; Mover según mejor estado
     
     ((pt-util:consulta-contador) ; Agotado el tiempo de búsqueda
      (envio:mensaje "Agotado el tiempo de búsqueda")
      (estado-dir-ini est-sol))                ; Mover según primero frente
     
     (t                           ; Encontrado camino
      (envio:mensaje "Encontrado camino")
      (print-camino (estado-camino est-sol))    ; Imprimir camino hallado
      (estado-dir-ini est-sol)))))              ; Mover según primero frente


;;
;; Definición de agentes presentes en este package
;; --------------------------------------------------------

(pt:defagente dpddba
  "Agente depredador por búsqueda en anchura"
  :turno #'dpddba
  :inicio #'dpddba-ini
  :tam-col t
  :mov-con t
  :for-cab 2
  :for-col 1
  :use '("CONTADOR" "SHELL" "GENERIC-SEARCH"))

(pt:defagente dpddbf
  "Agente depredador por búsqueda primero el mejor"
  :turno #'dpdda
  :inicio #'dpddbf-ini
  :tam-col t
  :mov-con t
  :for-cab 2
  :for-col 1
  :use '("CONTADOR" "SHELL" "GENERIC-SEARCH" "A_STAR"))

(pt:defagente dpdda
  "Agente depredador por búsqueda de camino en A*"
  :turno #'dpdda
  :inicio #'dpdda-ini
  :tam-col t
  :mov-con t
  :for-cab 2
  :for-col 1
  :use '("CONTADOR" "SHELL" "GENERIC-SEARCH" "A_STAR"))
